#ifndef _ANCHOR_3D_RANGE_GEN_
#define _ANCHOR_3D_RANGE_GEN_

#include <vector>
#include <iostream>
#include "Utils_Process.h"
#include <Eigen/CXX11/Tensor>

template<typename T>
class Anchor3DRangeGenerator{
private:
    const Vec2D<T> ranges;
    const Vec<int> scales;
    const Vec2D<T> sizes;
    const Vec<T> rotations;
    const Vec<T> custom_values;
    const bool reshape_out = true;
    const bool align_corner = false;
    const bool size_per_range = true;

    Tensor_R<T, 6> anchors_single_range(
        Vec<int> feature_size,
        const Vec<T>& anchor_range,
        const int& scale,
        Vec<T> in_sizes = Vec<T>{2.5981, 0.866, 1.0},
        const Vec<T>& in_rotations = Vec<T>{0.0, 1.5707963}
    );

    Tensor_R<T, 6> single_level_grid_anchors(
        Vec<int> feature_size,
        const int& scale
    );

public:
    Anchor3DRangeGenerator(
        const Vec2D<T> ranges,
        const Vec<int> scales = Vec<int>{1},
        const Vec2D<T> sizes = Vec2D<T>{{3.9, 1.6, 1.56}},
        const Vec<T> rotations = Vec<T>{0.0, 1.5707963},
        const bool reshape_out = true,
        const bool align_corner = false,
        const bool size_per_range = true)
        : ranges{ranges}, scales{scales}, sizes{sizes},
          rotations{rotations}, reshape_out{reshape_out},
          align_corner{align_corner}, size_per_range{size_per_range} {}

    Vec<Tensor_R<T, 2>> grid_anchors(
        Vec2D<int> featmap_sizes
    );
};

template<typename T>
Vec<Tensor_R<T, 2>> Anchor3DRangeGenerator<T>::grid_anchors(
    Vec2D<int> featmap_sizes)
{
    Vec<Tensor_R<T, 2>> multi_level_anchors;
    for(int i = 0; i < this->scales.size(); i++){
        Tensor_R<T, 6> anchors =
            this->single_level_grid_anchors(featmap_sizes[i], this->scales[i]);

        Tensor_R<T, 2> anchors_reshaped;
        if(this->reshape_out){
            auto dims = anchors.dimensions();
            int other_dim = 1;
            for(int j = 0; j < dims.size()-1; j++)
                other_dim *= dims[j];
            anchors_reshaped = anchors.reshape(Arr<Eigen::Index, 2>{other_dim, dims[dims.size()-1]});
        }

        multi_level_anchors.push_back(anchors_reshaped);
    }
    return multi_level_anchors;
}

template<typename T>
Tensor_R<T, 6> Anchor3DRangeGenerator<T>::single_level_grid_anchors(
    Vec<int> feature_size,
    const int& scale)
{
    Vec<Tensor_R<T, 6>> mr_anchors;
    for(int i = 0; i < this->ranges.size(); i++)
        mr_anchors.push_back(
            this->anchors_single_range(
                feature_size,
                this->ranges[i],
                scale,
                this->sizes[i],
                this->rotations
            )
        );

    //FIXME: I assume ranges.size() == 4 or this simple
    // concatenation gets so messy as far as I know
    Tensor_R<T, 6> mr_anchors_stacked =
        mr_anchors[0].concatenate(mr_anchors[1], 3).eval()
                     .concatenate(mr_anchors[2], 3).eval()
                     .concatenate(mr_anchors[3], 3).eval();

    return mr_anchors_stacked;
}

template<typename T>
Tensor_R<T, 6> Anchor3DRangeGenerator<T>::anchors_single_range(
    Vec<int> feature_size,
    const Vec<T>& anchor_range,
    const int& scale,
    Vec<T> in_sizes,
    const Vec<T>& in_rotations)
{
    if(feature_size.size() == 2)
        feature_size.insert(feature_size.begin(), 1);

    Eigen::VectorX<T> z_centers = Eigen::VectorX<T>::LinSpaced(feature_size[0]+1, anchor_range[2], anchor_range[5]);
    Eigen::VectorX<T> y_centers = Eigen::VectorX<T>::LinSpaced(feature_size[1]+1, anchor_range[1], anchor_range[4]);
    Eigen::VectorX<T> x_centers = Eigen::VectorX<T>::LinSpaced(feature_size[2]+1, anchor_range[0], anchor_range[3]);

    Eigen::Map<Eigen::VectorX<T>> sizes(in_sizes.data(), in_sizes.size());
    sizes = (sizes*scale).eval();
    Eigen::Map<const Eigen::VectorX<T>> rotations(in_rotations.data(), in_rotations.size());

    if(!this->align_corner){
        T z_shift = (z_centers[1]-z_centers[0]) / 2;
        T y_shift = (y_centers[1]-y_centers[0]) / 2;
        T x_shift = (x_centers[1]-x_centers[0]) / 2;

        z_centers = (z_centers.array()+z_shift).eval();
        y_centers = (y_centers.array()+y_shift).eval();
        x_centers = (x_centers.array()+x_shift).eval();
    }
    // Remove the last element
    x_centers.conservativeResize(x_centers.size()-1);
    y_centers.conservativeResize(y_centers.size()-1);
    z_centers.conservativeResize(z_centers.size()-1);

    Vec<Tensor_R<T, 4>> rets = create_4D_grid<T>(x_centers, y_centers, z_centers, rotations);

    Vec<Tensor_R<T, 6>> rets_reshaped;
    for(Tensor_R<T, 4>& ret : rets)
        rets_reshaped.push_back(unsqueeze<T, 5>(unsqueeze<T, 4>(ret, 2), 5));

    Tensor_R<T, 1> sizes_stacked(3);
    sizes_stacked(0) = sizes[0];
    sizes_stacked(1) = sizes[1];
    sizes_stacked(2) = sizes[2];

    auto ret_dim = rets_reshaped[0].dimensions();
    ret_dim[5] = 3;
    Tensor_R<T, 6> sizes_rep(ret_dim);
    for(int i = 0; i < ret_dim[0]; i++)
        for(int j = 0; j < ret_dim[1]; j++)
            for(int k = 0; k < ret_dim[4]; k++)
                sizes_rep.chip(i, 0)
                         .chip(j, 0)
                         .chip(0, 0)
                         .chip(0, 0)
                         .chip(k, 0) = sizes_stacked;

    rets_reshaped.insert(rets_reshaped.begin()+3, sizes_rep);

    Tensor_R<T, 6> ret(ret_dim[0], ret_dim[1],
                       ret_dim[2], ret_dim[3],
                       ret_dim[4], 9);
    // Stacking up + permuting
    //FIXME: If there are more than 6, this is not gonna work
    ret = rets_reshaped[0].concatenate(rets_reshaped[1], 5).eval()
                          .concatenate(rets_reshaped[2], 5).eval()
                          .concatenate(rets_reshaped[3], 5).eval()
                          .concatenate(rets_reshaped[4], 5).eval();

    static constexpr Arr<int, 6> perm_order{2, 1, 0, 3, 4, 5};
    Tensor_R<T, 6> ret_permuted = ret.shuffle(perm_order);
    Tensor_R<T, 6> ret_final(1, ret_dim[1], ret_dim[0], 1, 2, 9);
    ret_final.setZero();
    for(int i = 0; i < 7; i++)
        ret_final.chip(i, 5) = ret_permuted.chip(i, 5);
    return ret_final;
}

#endif // _ANCHOR_3D_RANGE_GEN_
